// cBackground.cpp

#include "cBackground.h"

#include <GL/gl.h>

#include <cstdio>
#include <memory.h>
#include <stdlib.h>

#include <iostream>
using std::cout;
using std::endl;

#include "psi3d/macros.h"

extern unsigned char* loadTGA(const char *fname, int *w, int* h, int* bpp);

float nm_to_R(int nm) {
    if (nm <= 390) return 1.0f;
    if (nm >= 440 && nm <= 510) return 0.0f;
    if (nm >= 580) return 1.0f;
    if (nm < 440) return (float(nm - 390) / float(440 - 390));
    return (float(nm - 510) / float(580 - 510));
}

float nm_to_G(int nm) {
    if (nm <= 440) return 0.0f;
    if (nm >= 640) return 1.0f;
    if (nm >= 490 && nm <= 580) return 1.0f;
    if (nm < 490) return (float(nm - 440) / float(490 - 440));
    return (float(nm - 580) / float(640 - 580));
}

float nm_to_B(int nm) {
    if (nm < 490) return 1.0f;
    if (nm > 510) return 0.0f;
    return (float(nm - 490) / float(510 - 490));
}

enum Texids {
    T_ICE, T_HAZE, T_MOUNTAIN, T_SUN, T_MOON, T_EARTH, T_LAVOS
};

cDomeBackground::cDomeBackground() {

    struct Texture {
        const char* filename;
        bool soft, repeat;
        int id;
    } textures[] = {
        { "data/planet/background/ice.tga", false, true, T_ICE},
        { "data/planet/background/haze.tga", true, false, T_HAZE},
        { "data/planet/background/mountain.tga", true, false, T_MOUNTAIN},
        { "data/planet/background/sun.tga", true, false, T_SUN},
        { "data/planet/background/moon.tga", true, false, T_MOON},
        { "data/planet/background/earth.tga", true, false, T_EARTH},
        { "data/planet/background/lavos.tga", true, false, T_LAVOS}
    };

    loopi(7) {
        int w, h, bpp;
        unsigned char* texels = loadTGA(textures[i].filename, &w, &h, &bpp);
        bool soft = textures[i].soft;
        bool repeat = textures[i].repeat;
        unsigned int texname;
        glBindTexture2D(texname, true, soft, repeat, repeat, w, h, bpp, texels);
        int id = textures[i].id;
        mTextures[id] = texname;
    }
}

void cDomeBackground::drawBackground(float hour) {
    // Push random seed because the current seed is going to be
    // replaced by a (deterministic) constant value for clouds.
    unsigned int seed = rand();

    mHour = hour;
    //float timescale = 123000;
    //hour = rand()%24;

    // hour = [0,24]
    // => [-90,   0, 90, 180, 270]
    // => [ -1,   0,  1,   0,  -1] = sin
    // => [  0, 0.5,  1, 0.5,   0] = s brightness sine-wave.
    float t = ((hour * 360.0 / 24.0) - 90.0); // = [-90,+270]
    float s = sin(t * 0.017453)*0.5 + 0.5; // = [0,1]

    // Set gradient colors according to brightness sine-wave.
    vector_set(mUpColor, 0.0, 0.0, s * 0.8 + 0.2);
    vector_set(mLoColor, s * 0.95 + 0.05, s * 0.95 + 0.05, s * 0.95 + 0.05);

    //vector_set(mBtColor, 0.4 * s + 0.1, 0.4 * s + 0.1, 0.4 * s + 0.1);
    vector_set(mBtColor, 0.4 * s + 0.1, 0.2 * s + 0.1, 0.1 * s + 0.1);
    //vector_set(mBtColor, 0.6 * s + 0.1, 0.6 * s + 0.1, 0.9 * s + 0.1);
    //vector_set(mBtColor, 0.44 * s + 0.1, 0.44 * s + 0.1, 0.46 * s + 0.1);
    //vector_set(mBtColor, 0.24 * s + 0.1, 0.44 * s + 0.1, 0.26 * s + 0.1);

    if (1) {
        float p[] = {700, 100, 0, 1};
        float a[] = {0.3, 0.3, 0.3, 1};
        glLightfv(GL_LIGHT0, GL_POSITION, p);
        glLightfv(GL_LIGHT0, GL_AMBIENT, a);
        glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
        glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0001);
        glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.000001);
        glEnable(GL_NORMALIZE);
    }


    GLint mode;
    glGetIntegerv(GL_POLYGON_MODE, &mode);
    if (mode == GL_FILL) {
        drawUpperDome();
        drawOrbit(hour);
        drawClouds();
        drawMountains();
        drawLowerDome();
        drawGround();
    } else {
        //drawUpperDome();
        //drawMountains();
        //drawLowerDome();
    }

    glFogfv(GL_FOG_COLOR, mBtColor);

    // Pop a seed.
    srand(seed);
}

void cDomeBackground::drawGround() {
    //return;
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_LIGHTING);
        glEnable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glBindTexture(GL_TEXTURE_2D, mTextures[T_ICE]);

        // Get current Camera-Matrix.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        // Position from 4th Col.
        float p[3];
        loop3i(p[i] = m[12 + i]);
        // Reset Position on 4th Col.
        loop3i(m[12 + i] = 0);
        // Apply transpose-inverse to position.
        float n[16];
        memcpy(n, m, sizeof (float) *16);
        matrix_transpose(m);
        matrix_apply2(m, p);
        //cout << "height over ground: " << p[1] << "\n";
        //cout << "Position: " << p[0] << " " << p[2] << "\n";

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(n); // Transpose to undo previous transpose.
            glTranslatef(0, p[Y], 0);

            float color2[3] = {0.4, 0.2, 0.1};
            vector_cpy(color2, mBtColor);

            const float tdiv = 5;
            const float radius = 200;

            float dx = -p[0] / tdiv;
            float dz = -p[2] / tdiv;

            glBegin(GL_TRIANGLE_FAN);
            glColor4f(mBtColor[0], mBtColor[1], mBtColor[2], 1.0f);
            glNormal3f(0, 1, 0);
            glTexCoord2f(dx, dz);
            glVertex3f(0.0, 0.0, 0.0);
            glColor4f(mBtColor[0], mBtColor[1], mBtColor[2], 0.0f);
            float e = 20.0 * M_PI / 180.0f;

            loopi(19) {
                glNormal3f(0, 1, 0);
                glTexCoord3f(cos(-i * e) * radius / tdiv + dx, sin(-i * e) * radius / tdiv + dz, 0);
                glVertex3f(cos(-i * e) * radius, 0, sin(-i * e) * radius);
            } // loopi 19
            glEnd();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawUpperDome() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);
        glDisable(GL_FOG);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100 * 0.25, 100);

            int vsteps = 8;
            double vf = 0.5 * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            float color1[3];
            vector_cpy(color1, mUpColor);

            float color2[3];
            vector_cpy(color2, mLoColor);

            float slope[3];
            vector_set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            vector_cpy(color2, color1);
            vector_add(color2, color2, slope);

            float s1 = sin(0);
            float c1 = cos(0);
            for (int i = 1; i <= vsteps; i++) {
                float s2 = sin(i * vf);
                float c2 = cos(i * vf);
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor3fv(color1);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    glColor3fv(color2);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    h += hf;
                } // for hsteps
                glEnd();
                vector_cpy(color1, color2);
                vector_add(color2, color2, slope);
                s1 = s2;
                c1 = c2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawLowerDome() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glEnable(GL_CULL_FACE);
        glDisable(GL_FOG);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100 * 0.1, 100);

            int vsteps = 8;
            double vf = 0.5f * M_PI / (double) vsteps;

            int hsteps = 18;
            double hf = 2.0 * M_PI / (double) hsteps;
            double h = 0;

            const float f = 0.5;
            float color2[3] = {0.4, 0.2, 0.1};
            vector_cpy(color2, mBtColor);
            float color1[3] = {color2[0] * f, color2[1] * f, color2[2] * f};

            float tmp[3];
            vector_cpy(tmp, color1);
            vector_cpy(color1, color2);
            vector_cpy(color2, tmp);

            float slope[3];
            vector_set3i(slope, (color2[i] - color1[i]) / (float) vsteps);

            vector_cpy(color2, color1);
            vector_add(color2, color2, slope);

            float s1 = -sin(0);
            float c1 = -cos(0);
            for (int i = 1; i <= vsteps; i++) {
                float s2 = -sin(i * vf);
                float c2 = -cos(i * vf);
                glBegin(GL_TRIANGLE_STRIP);
                for (int j = 0; j <= hsteps; j++) {
                    float s = sin(h);
                    float c = cos(h);
                    glColor3fv(color2);
                    glNormal3f(s2*c, c2, s2 * s);
                    glVertex3f(s2*c, c2, s2 * s);
                    glColor3fv(color1);
                    glNormal3f(s1*c, c1, s1 * s);
                    glVertex3f(s1*c, c1, s1 * s);
                    h += hf;
                };
                glEnd();
                vector_cpy(color1, color2);
                vector_add(color2, color2, slope);
                s1 = s2;
                c1 = c2;
            } // for vsteps
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawClouds() {
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glDisable(GL_FOG);
        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, mTextures[T_HAZE]);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        //m[12 + 0] = m[12 + 2] = 0;

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            glScalef(100, 100 * 0.125, 100);

            float windspeed = 25;
            float trad = windspeed * 24.0f * (mHour / 24.00f) * 2 * M_PI;
            glRotatef(trad / 0.017453f, 1, 0.1, 1);

            //cout << mHour << endl;
            srand(12421);
            int n = 16;
            float density = 3.0f / float(1.0f + n);
            for (int i = 0; i < n; i++) {
                //float xrad = -(rand() % 157) * 0.01f;
                float xrad = -(157 - (rand() % 314)) * 0.01f;
                float yrad = (rand() % 628) * 0.01f;

                /*
                float xaxis[] = { 1,0,0 };
                float qx[4];
                quat_rotaxis(qx, xrad, xaxis);

                float yaxis[] = { 0,1,0 };
                float qy[4];
                quat_rotaxis(qy, yrad, yaxis);

                float qxy[4];
                quat_mul(qxy, qx, qy);

                float x[] = {1,0,0};
                float y[] = {0,1,0};
                float z[] = {0,0,1};
                quat_apply(x, qxy, x);
                quat_apply(y, qxy, y);
                quat_apply(z, qxy, z);

                glColor4f(1,1,1,1);
                glVertex3f(1*x[0]+z[0], 1+z[1], 0+z[2]);
                glVertex3f(0, 1, 0);
                glVertex3f(0, 0, 0);
                glVertex3f(1, 0, 0);
                glEnd();
                 */

                glPushMatrix();
                {
                    glRotatef(yrad / 0.017454, 0, 1, 0);
                    glRotatef(xrad / 0.017454, 1, 0, 0);
                    glTranslatef(0, 0, 1.7);
                    glColor4f(1, 1, 1, density);
                    glBegin(GL_QUADS);
                    glTexCoord2f(1, 1);
                    glVertex3f(+1, +1, 0);
                    glTexCoord2f(0, 1);
                    glVertex3f(-1, +1, 0);
                    glTexCoord2f(0, 0);
                    glVertex3f(-1, -1, 0);
                    glTexCoord2f(1, 0);
                    glVertex3f(+1, -1, 0);
                    glEnd();
                }
                glPopMatrix();
            }

        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawMountains() {
    srand(762381);
    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        glEnable(GL_FOG);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glDisable(GL_CULL_FACE);
        glEnable(GL_TEXTURE_2D);

        glBindTexture(GL_TEXTURE_2D, mTextures[T_MOUNTAIN]);
        glColor3fv(mBtColor);

        // Load current Camera Matrix and
        // set Position (4th Column) = 0.
        float m[16];
        glGetFloatv(GL_MODELVIEW_MATRIX, m);
        loop3i(m[12 + i] = 0);
        //m[12 + 0] = m[12 + 2] = 0;

        glPushMatrix();
        {
            int n = 30;
            float f = 1.0;
            for (int i = 0; i < n; i++) {
                glLoadIdentity();
                glMultMatrixf(m);
                glScalef(90, 100 * 0.2, 90);

                float w = (0.3 + (rand() % 100) * 0.005f) * 0.5 * f;
                float rad = 0.01f * (rand() % 628);
                glRotatef(rad / 0.017453, 0, 1, 0);
                glTranslatef(0, 0, 1);

                float u = 1.2 * f;
                glBegin(GL_QUADS);
                glTexCoord2f(0, 1);
                glVertex3f(-w, u, 1);
                glTexCoord2f(0, 0);
                glVertex3f(-w, -0.01, 1);
                glTexCoord2f(1, 0);
                glVertex3f(+w, -0.01, 1);
                glTexCoord2f(1, 1);
                glVertex3f(+w, u, 1);
                glEnd();

                f = f * 0.95f;
            }
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cDomeBackground::drawOrbit(float hour) {
    float m[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, m);

    // 4th Row = 0.
    m[0 * 4 + 3] = m[1 * 4 + 3] = m[2 * 4 + 3] = 0.0f;

    // 4th Col = 0.
    m[3 * 4 + 0] = m[3 * 4 + 1] = m[3 * 4 + 2] = 0.0f;
    m[3 * 4 + 3] = 1.0f;

    glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
    {
        //glDisable(GL_CULL_FACE);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_FOG);
        glDisable(GL_LIGHTING);

        glPushMatrix();
        {
            glLoadIdentity();
            glMultMatrixf(m);
            //glScalef(1,0.3f,1);
            
            const float scale = 0.1 * 2400.0f;
            float sun = 7 * 80 * 0.1;
            //const float corona = 1.2f;
            float moon = sun * 0.2;

            static float angle = 0;
            angle -= 1;

            angle = 180 - ((360 / 24.00f) * hour + 90);
            //  angle = 180-((360 / 24.00f) * 20 + 90);

            //angle = 0;

            // Sun travels: dawn 0600:East, 1200:South, dusk 1800:West (2400:North -> other side of the globe).
            // 0000/2400: 0 degree, 0600: 90 degree, 1200: 180 degree, 1800: 270 degree, ...

            // Moon is traveling about 180 degree behind/ahead.

            // Well, that is, if you're down-under then east, north, west (not east, south, west).

            glRotatef(-90, 0, 1, 0);
            //	glRotatef(45+7.5, 0, 0, 1); // Longitude? Aquator Angle Offset -> Germany. and for example Poland, GB and Canada.
            glRotatef(55, 0, 0, 1);
            glRotatef(0, 0, 0, 1); // Aquator.

            const int detail = 7;

            if (!true) { // atmospheric scattering
                glPushMatrix();
                {
                    glRotatef(angle, 1, 0, 0);
                    float dist = 0.005;
                    glBegin(GL_TRIANGLE_FAN);
                    {
                        glColor4f(1, 1, 1, 0.99);
                        glVertex3f(0, 0, scale);
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, scale * dist);
                            alpha += inc;
                        }
                    }
                    glEnd();
                    glBegin(GL_TRIANGLE_STRIP);
                    {
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, -scale * dist);
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, +scale * dist);
                            alpha -= inc;
                        }
                    }
                    glEnd();
                    glBegin(GL_TRIANGLE_FAN);
                    {
                        glColor4f(1, 1, 1, 0.3);
                        glVertex3f(0, 0, -scale);
                        float alpha = 0;
                        const float inc = 2.0f * M_PI / float(detail - 1);
                        const float r = scale;

                        loopi(detail) {
                            glColor4f(0, 0, 0.3, 0.1);
                            glVertex3f(sin(alpha) * r, cos(alpha) * r, -scale * dist);
                            alpha -= inc;
                        }
                    }
                    glEnd();
                }
                glPopMatrix();
            } // atmospheric scattering

            glPushMatrix();
            //if (hour > 6.50 && hour < 17.50)
            {
                glRotatef(angle, 1, 0, 0);

                glTranslatef(0, 0, scale);
                glColor4f(1, 0.95, 0.9, 0.7);
                //glDisk(detail, sun);
                // Sun
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, mTextures[T_SUN]);
                glBegin(GL_QUADS);
                glTexCoord2f(0, 0);
                glVertex3f(-sun, -sun, 0);
                glTexCoord2f(0, 1);
                glVertex3f(-sun, +sun, 0);
                glTexCoord2f(1, 1);
                glVertex3f(+sun, +sun, 0);
                glTexCoord2f(1, 0);
                glVertex3f(+sun, -sun, 0);
                glEnd();
            }
            glPopMatrix();

            glPushMatrix();
            //if (hour > 18.50 || hour < 5.50)
            {
                glRotatef(angle + 180, 1, 0, 0);

                glTranslatef(0, 0, scale);
                glColor4f(0.6, 0.6, 0.7, 0.94);
                //glDisk(detail, moon);
                // Moon
                glEnable(GL_TEXTURE_2D);
                glBindTexture(GL_TEXTURE_2D, mTextures[T_MOON]);
                glBegin(GL_QUADS);
                glTexCoord2f(0, 0);
                glVertex3f(-moon, -moon, 0);
                glTexCoord2f(0, 1);
                glVertex3f(-moon, +moon, 0);
                glTexCoord2f(1, 1);
                glVertex3f(+moon, +moon, 0);
                glTexCoord2f(1, 0);
                glVertex3f(+moon, -moon, 0);
                glEnd();
            }
            glPopMatrix();

        }
        glPopMatrix();
    }
    glPopAttrib();
}
