/**
 * File:     cBackground.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on March 28, 2008, 21:25 PM (1999)
 *
 * Virtual Horizon with Sky and Ground
 * (SkySphere / SkyBox).
 */

#ifndef _CBACKGROUND_H
#define _CBACKGROUND_H

#include <map>

class cBackground;
class cBlankBackground;
class cDomeBackground;

/**
 * Abstract Base-Class for background painting.
 * Only method will be called on begin of frame painting.
 */
class cBackground {
public:
    /**
     * Draws the worlds background for a given time.
     */
    virtual void drawBackground(float hour) = 0;
};

/**
 * Just a Background which leaves the background empty
 * so that only the ClearColored background stays.
 */
class cBlankBackground : public cBackground {
public:
    void drawBackground(float hour = 12.00f) {
    }
};

/**
 * A background in form of a SkySphere just like
 * those virtual horizons found in aviation instruments.
 */
class cDomeBackground : public cBackground {
public:
    float mHour;
    std::map<int,unsigned int> mTextures;
    float mUpColor[3];
    float mLoColor[3];
    float mBtColor[3];
public:
    cDomeBackground();
    void drawBackground(float hour = 12.00f); // Calls other draw*()
    void drawUpperDome();
    void drawLowerDome();
    void drawGround();
    void drawClouds();
    void drawMountains();
    void drawOrbit(float hour);
};

#endif

