// cBuilding.cpp

#include "cBuilding.h"

#include "psi3d/macros.h"

#include <GL/gl.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>

#include <iostream>
using std::cout;
using std::endl;


int cBuilding::sInstances = 0;
std::map<int,long> cBuilding::sTextures;

cBuilding::cBuilding(int x, int y, int z, int rooms_x, int rooms_y, int rooms_z) {
    sInstances++;
    if (sInstances == 1) {
        unsigned int texname;

        {
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/urban/scraper/concrete.tga", &w, &h, &bpp);
            //cout << w << " " << h << " " << bpp << endl;
            glBindTexture2D(texname, true, false, true, true, w, h, bpp, texels);
            delete texels;
            sTextures[3] = texname;
        }
        {
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/urban/scraper/scraper.tga", &w, &h, &bpp);
            //cout << w << " " << h << " " << bpp << endl;
            glBindTexture2D(texname, true, false, true, true, w, h, bpp, texels);
            delete texels;
            sTextures[0] = texname;
            sTextures[1] = texname;
            sTextures[2] = texname;
        }
    }
    collideable = new rCollideable;
    damageable = new rDamageable;
    addRole(COLLIDEABLE);
    addRole(DAMAGEABLE);
    addRole(BUILDING);
    quat_set(traceable->ori, 0, 0, 0, 1);
    mRooms[0] = 3 * rooms_x;
    mRooms[1] = 3 * rooms_y;
    mRooms[2] = 3 * rooms_z;
    vector_set(traceable->pos, x * 3 + mRooms[0]*0.5f, y * 3, z * 3 + mRooms[2]*0.5f);
    this->traceable->radius = sqrt(0.25f * mRooms[0] * mRooms[0] + 0.25f * mRooms[1] * mRooms[1] + 0.25f * mRooms[2] * mRooms[2]);
    mDirtybase = true;
}

void cBuilding::damageByParticle(float* localpos, float damage, cObject* enactor) {
    if (damage > 0) {
        this->damageable->hp[rDamageable::BODY] -= damage;
        if (this->damageable->hp[rDamageable::BODY] <= 0 && !hasRole(DEAD)) {
            this->remRole(DAMAGEABLE);
            this->addRole(DEAD);
            cout << "cBuilding::damageByParticle(): DEAD\n";
            mExplosion.fire(0);
        }
    }
}

float cBuilding::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
    float localpos_[3];

    {
        float ori_inv[4];
        quat_cpy(ori_inv, traceable->ori);
        quat_conj(ori_inv);

        vector_cpy(localpos_, worldpos);
        vector_sub(localpos_, localpos_, traceable->pos);
        quat_apply(localpos_, ori_inv, localpos_);
    }

    float* w = this->mRooms;
    float mins[3] = {-w[0]*0.5 - radius, -0.0 - radius, -w[2]*0.5 - radius};
    float maxs[3] = {+w[0]*0.5 + radius, +w[1] + radius, +w[2]*0.5 + radius};
    float localprj[3];

    //cout << endl;
    //cout << mins[0] << " " << mins[1] << " " << mins[2] << endl;
    //cout << localpos[0] << " " << localpos[1] << " " << localpos[2] << endl;
    //cout << maxs[0] << " " << maxs[1] << " " << maxs[2] << endl;

    float depth = 0;
    depth = cParticle::constraintParticleByAABB(localpos_, mins, maxs, localprj);

    //cout << depth << endl;

    if (depth <= 0) return 0;

    if (localpos != NULL) vector_cpy(localpos, localprj);

    {
        quat_apply(worldpos, traceable->ori, localprj);
        vector_add(worldpos, worldpos, traceable->pos);
    }

    return depth;
}

void cBuilding::onSpawn() {
    mExplosion.mOwner = this;
}

void cBuilding::animate(float spf) {
    mExplosion.animate(spf);
}

void cBuilding::transform() {
    if (mDirtybase) {
        mDirtybase = false;
        glPushMatrix();
        {
            float* p = this->traceable->pos.data();
            float r[] = { 0,0,0 };
            float* w = this->mRooms;
            glTranslatef(p[0], p[1], p[2]);
            if (r[1] < -0.00001f || r[1] > 0.00001f) glRotatef(r[1] * 0.017453f, 0, 1, 0);
            glTranslatef(0, 0.125f * 3.5f * w[1], 0);
            glGetFloatv(GL_MODELVIEW_MATRIX, mBase);
            mExplosion.mBasefv = mBase;
        }
        glPopMatrix();
    }
    mExplosion.transform();
}

void cBuilding::drawSolid() {
    mExplosion.drawSolid();

    float* p = this->traceable->pos.data();
    float r[] = { 0,0,0 };
    float* w = this->mRooms;

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glEnable(GL_CULL_FACE);
        glEnable(GL_TEXTURE_2D);
        long roof = sTextures[3];
        long wall = 0;
        if (this->damageable->hp[rDamageable::BODY] <= 0) {
            glColor3f(0.6f, 0.5f, 0.4f);
            wall = sTextures[2];
        } else if (this->damageable->hp[rDamageable::BODY] <= 50) {
            glColor3f(0.6f, 0.7f, 0.8f);
            wall = sTextures[1];
        } else {
            glColor3f(0.7f, 0.8f, 0.9f);
            wall = sTextures[0];
        }

        float hp = this->damageable->hp[rDamageable::BODY] * 0.01;
        hp = hp < 0 ? 0 : hp;
        float d = 0.5f + hp * 0.5f;
        glColor3f(d, d, d);

        glPushMatrix();
        {
            if (r[1] < -0.00001f || r[1] > 0.00001f) glRotatef(r[1] * 0.017453f, 0, 1, 0);
            glTranslatef(p[0] - w[0]*0.5f, p[1], p[2] - w[2]*0.5f);
            //glBlockFlat(w[0], w[1], w[2], wall, wall, wall, wall, roof, roof);
            const float n2 = 0.707107f; // 1.0f / sqrt(2);
            const float n3 = 0.577350f; // 1.0f / sqrt(3);
            glBindTexture(GL_TEXTURE_2D, roof);
            glBegin(GL_QUADS);
            {
                glNormal3f(0, 1, 0);

                glTexCoord2f(0, 0);
                glVertex3f(0, w[1], 0);

                glTexCoord2f(0, w[2]);
                glVertex3f(0, w[1], w[2]);

                glTexCoord2f(w[0], w[2]);
                glVertex3f(w[0], w[1], w[2]);

                glTexCoord2f(w[0], 0);
                glVertex3f(w[0], w[1], 0);
            }
            glEnd();
            glBindTexture(GL_TEXTURE_2D, wall);
            glBegin(GL_TRIANGLE_STRIP);
            {
                float tx = 0;
                float f = 1.0f / 9.0f * 2.0f;
                float ty = w[1] * f;

                glNormal3f(-n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, 0);

                glNormal3f(-n3, n3, -n3);
                glTexCoord2f(tx, ty);
                glVertex3f(0, w[1], 0);

                tx += w[0] * f;

                glNormal3f(+n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(w[0], 0, 0);

                glNormal3f(+n3, n3, -n3);
                glTexCoord2f(tx, ty);
                glVertex3f(w[0], w[1], 0);

                tx += w[2] * f;

                glNormal3f(+n2, 0, +n2);
                glTexCoord2f(tx, 0);
                glVertex3f(w[0], 0, w[2]);

                glNormal3f(+n3, n3, +n3);
                glTexCoord2f(tx, ty);
                glVertex3f(w[0], w[1], w[2]);

                tx += w[0] * f;

                glNormal3f(-n2, 0, +n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, w[2]);

                glNormal3f(-n3, n3, +n3);
                glTexCoord2f(tx, ty);
                glVertex3f(0, w[1], w[2]);

                tx += w[2] * f;

                glNormal3f(-n2, 0, -n2);
                glTexCoord2f(tx, 0);
                glVertex3f(0, 0, 0);

                glNormal3f(-n3, n3, -n3);
                glTexCoord2f(tx, ty);
                glVertex3f(0, w[1], 0);
            }
            glEnd();

            if (!true) glNormAxis();
        }
        glPopMatrix();
    }
    glPopAttrib();
}

void cBuilding::drawEffect() {
    mExplosion.drawEffect();
}


// ###############################################


#define glTexturedSquare(bind) \
{ \
    glBindTexture(GL_TEXTURE_2D, (unsigned int) bind); \
    glBegin(GL_QUADS); \
    glTexCoord2f(0, 0); \
    glVertex3f(-0.5f, -0.5f, 0); \
    glTexCoord2f(1, 0); \
    glVertex3f(+0.5f, -0.5f, 0); \
    glTexCoord2f(1, 1); \
    glVertex3f(+0.5f, +0.5f, 0); \
    glTexCoord2f(0, 1); \
    glVertex3f(-0.5f, +0.5f, 0); \
    glEnd(); \
}

#include "cParticle.h"

int cScatterfield::sInstances = 0;
std::map<int,long> cScatterfield::sTextures;

cScatterfield::cScatterfield(float x, float y, float z, float radius, float density) {
    sInstances++;
    if (sInstances == 1) {
        {
            unsigned int texid;
            int w, h, bpp;
            unsigned char* texels = loadTGA("data/planet/background/mountain.tga", &w, &h, &bpp);
            glBindTexture2D(texid, true, true, false, false, w, h, bpp, texels);
            delete texels;
            sTextures[0] = texid;
        }
    }

    vector_set(traceable->pos, x, y, z);

    int amount = density * 4 * radius * M_PI;
    for (int i = 0; i < amount; i++) {
        float x_ = x + (rand() % 1000) * 0.001 * 2 * radius - radius;
        float y_ = y;
        float z_ = z + (rand() % 1000) * 0.001 * 2 * radius - radius;
        float r_ = (rand()%360);
        cParticle* p = new cParticle();
        assert(p != NULL);
        vector_set(p->pos, x_, y_, z_);
        p->spawn = r_;
        mDecals.push_back(p);
    }
}

void cScatterfield::drawEffect() {
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_CULL_FACE);
        glEnable(GL_TEXTURE_2D);
        glColor4f(0.3, 0.7, 0.3, 1);
        foreachNoInc(i, mDecals) {
            cParticle* p = *i++;
            glPushMatrix();
            {
                glTranslatef(p->pos[0], p->pos[1], p->pos[2]);
                glRotatef(p->spawn, 0,1,0);
                //glAxis(0.9);
                glTexturedSquare(sTextures[0]);
            }
            glPopMatrix();
        }
    }
    glPopAttrib();
}


// ###############################################


int cTile::sInstances = 0;
std::map<int,long> cTile::sTextures;
//long cTile::sDisplaylists[cTile::MAX_KINDS * 2] = {0,};

cTile::cTile(int x, int y, int z, int kind) {
    sInstances++;
    if (sInstances == 1) {
        unsigned char* texels = NULL;
        unsigned int texname;
        int w, h, bpp;

        texels = loadTGA("data/urban/street/junction.tga", &w, &h, &bpp);
        glBindTexture2D(texname, true, false, true, true, w, h, bpp, texels);
        sTextures[KIND_ROAD_JUNKTION] = texname;

        texels = loadTGA("data/urban/street/ns_road.tga", &w, &h, &bpp);
        glBindTexture2D(texname, true, false, true, true, w, h, bpp, texels);
        sTextures[KIND_ROAD_NORTHSOUTH] = texname;

        texels = loadTGA("data/urban/street/we_road.tga", &w, &h, &bpp);
        glBindTexture2D(texname, true, false, true, true, w, h, bpp, texels);
        sTextures[KIND_ROAD_EASTWEST] = texname;
    }
    if (!true && sInstances == 1) {
        // First one initializes
        unsigned int texname;
        const int WIDTH = 8;
        const int HEIGHT = 8;
        const int DEPTH = 8;
        unsigned char texels[WIDTH * HEIGHT * DEPTH * 3];
        float s = 3;
        generateTexture3D(WIDTH, HEIGHT, DEPTH, texels, deltaFunc_NoisyPlasma, ColorFunc_Delta, ColorFunc_Delta, ColorFunc_Delta, 1 * s, 1 * s, 1 * s);
        //glBindTexture3D(texname, true, true, true, true, true, WIDTH, HEIGHT, DEPTH, texels);
        sTextures[0] = texname;
    }
    //    addRole(COLLIDEABLE); // !!
    vector_set(traceable->pos, x * 9.0f + 4.5f, y * 3.0f, z * 9.0f + 4.5f);
    mKind = kind;
}

void cTile::drawSolid() {
    float* p = this->traceable->pos.data();
    float r[] = { 0, 0, 0 };

    //glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        long ground = sTextures[mKind];
        glColor3f(1, 1, 1);
        glPushMatrix();
        {
            if (r[1] < -0.01f || r[1] > 0.01f) glRotatef(r[1], 0, 1, 0);
            glTranslatef(p[0] - 4.5f, p[1] + 0.08, p[2] - 4.5f);
            GLfloat p[][3] = {
                {0, 0, 0},
                {9, 0, 0},
                {0, 0, 9},
                {9, 0, 9}
            };
            const float a = 0.01;
            const float b = 0.99;
            GLfloat t[][2] = {
                {a, a},
                {b, a},
                {b, b},
                {a, b},
            };
            //glDisable(GL_CULL_FACE);
            glFrontFace(GL_CW);
            glEnable(GL_TEXTURE_2D);
            glColor4f(1, 1, 1, 0.9);
            glBindTexture(GL_TEXTURE_2D, ground);
            glBegin(GL_QUADS);
            glNormal3f(0, -1, 0);
            glVertex3fv2fv(p[0], t[1]);
            glVertex3fv2fv(p[1], t[0]);
            glVertex3fv2fv(p[3], t[3]);
            glVertex3fv2fv(p[2], t[2]);
            glEnd();
            if (!true) glNormAxis();
            glDisable(GL_LIGHTING);
            float h = 2;
            glColor3f(0.3, 0.3, 0.3);
            if (mKind == KIND_ROAD_JUNKTION && false) {
                glBegin(GL_LINES);
                glVertex3f(2, 0, 2);
                glVertex3f(2, h, 2);
                glVertex3f(7, 0, 2);
                glVertex3f(7, h, 2);
                glVertex3f(7, 0, 7);
                glVertex3f(7, h, 7);
                glVertex3f(2, 0, 7);
                glVertex3f(2, h, 7);
                glEnd();
            } else if (mKind == cTile::KIND_ROAD_EASTWEST) {
                glBegin(GL_LINE_STRIP);
                if ((int(traceable->pos[0] / 9) & 1) == 0) {
                    glVertex3f(5, 0, 2);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(5, h, 2);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(5, h + 0.2 * h, 2 + 0.3 * h);
                } else {
                    glVertex3f(5, 0, 7);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(5, h, 7);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(5, h + 0.2 * h, 7 - 0.3 * h);
                }
                glEnd();
            } else if (mKind == cTile::KIND_ROAD_NORTHSOUTH) {
                glBegin(GL_LINE_STRIP);
                if ((int(traceable->pos[2] / 9) & 1) == 0) {
                    glVertex3f(2, 0, 5);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(2, h, 5);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(2 + 0.3 * h, h + 0.2 * h, 5);
                } else {
                    glVertex3f(7, 0, 5);
                    glColor3f(0.7, 0.7, 0.6);
                    glVertex3f(7, h, 5);
                    glColor3f(1, 0.9, 0.7);
                    glVertex3f(7 - 0.3 * h, h + 0.2 * h, 5);
                }
                glEnd();
            }
        }
        glPopMatrix();
    }
    glPopAttrib();
}


// ###############################################

cPadmap::cPadmap(float x, float z) {
    vector_set(traceable->pos, x, 0, z);
    quat_set(traceable->ori, 0, 0, 0, 1);
    collideable = new rCollideable;
    addRole(COLLIDEABLE);
    float h[16 * 16] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

        0, 0, 0, 1, 1, 4, 4, 3, 3, 2, 2, 2, 2, 0, 0, 0,
        0, 0, 0, 1, 1, 4, 4, 3, 3, 2, 2, 2, 2, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 3, 2, 2, 2, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 2, 2, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 3, 3, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 3, 3, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 4, 4, 0, 0, 0, 0,

        0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 0, 0,
        0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 0, 0,

        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    };

    memcpy(mHeights, h, sizeof (float) *16 * 16);
}

void cPadmap::drawSolid() {
    //return; // !!

    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_TEXTURE_2D);
        //glDisable(GL_CULL_FACE);
        glColor4f(1, 1, 1, 1);

        int a = 16;
        int b = 16;

        /*
        glBegin(GL_TRIANGLE_STRIP);
        glVertex3f(0, 0, 0);
        glVertex3f(0, 0, b);
        glVertex3f(a, 0, 0);
        glVertex3f(a, 0, b);
        glEnd();
         */


        // Draw pad tops
        {
            glColor4f(0, 0, 1, 1);
            glNormal3f(0, 1, 0);
            int z = 0;
            for (int j = 0; j < b - 1; j += 2) {
                glBegin(GL_TRIANGLE_STRIP);
                int x = 0;
                for (int i = 0; i < a - 1; i += 2) {
                    bool stumble = (i == 0);
                    if (!stumble) {
                        stumble = (mHeights[(j + 0) * a + i - 1] != mHeights[(j + 0) * a + i])
                                || (mHeights[(j + 1) * a + i - 1] != mHeights[(j + 1) * a + i]);
                    }
                    if (stumble) {
                        glEnd();
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x, mHeights[(j + 0) * a + i], z);
                        glVertex3f(x, mHeights[(j + 1) * a + i], z + 1);
                    }
                    glVertex3f(x + 1, mHeights[(j + 0) * a + i + 1], z);
                    glVertex3f(x + 1, mHeights[(j + 1) * a + i + 1], z + 1);
                    x++;
                }
                glEnd();
                z++;
            }
        } // Draw pad tops

        // Draw east-west cliffs
        {
            glColor4f(1, 0, 1, 1);
            glNormal3f(0, 0, 1);
            int x = 0;
            for (int i = 0; i < a; i += 2) {
                int z = 1;
                for (int j = 2; j < b - 1; j += 2) {
                    bool stumble = (mHeights[(j - 1) * a + (i + 0)] != mHeights[(j + 0) * a + (i + 0)])
                            || (mHeights[(j - 1) * a + (i + 1)] != mHeights[(j + 0) * a + (i + 1)]);
                    if (stumble) {
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x + 1, mHeights[(j - 1) * a + i + 1], z);
                        glVertex3f(x + 0, mHeights[(j - 1) * a + i + 0], z);
                        glVertex3f(x + 1, mHeights[(j - 0) * a + i + 1], z);
                        glVertex3f(x + 0, mHeights[(j - 0) * a + i + 0], z);
                        glEnd();
                    }
                    z++;
                }
                x++;
            }
        } // Draw east-west cliffs

        // Draw north-south cliffs
        {
            glColor4f(0, 1, 1, 1);
            glNormal3f(1, 0, 0);
            int z = 0;
            for (int j = 0; j < b; j += 2) {
                int x = 1;
                for (int i = 2; i < a - 1; i += 2) {
                    bool stumble = (mHeights[(j + 0) * a + i - 1] != mHeights[(j + 0) * a + i])
                            || (mHeights[(j + 1) * a + i - 1] != mHeights[(j + 1) * a + i]);
                    if (stumble) {
                        glBegin(GL_TRIANGLE_STRIP);
                        glVertex3f(x, mHeights[(j + 0) * a + i - 1], z + 0);
                        glVertex3f(x, mHeights[(j + 1) * a + i - 1], z + 1);
                        glVertex3f(x, mHeights[(j + 0) * a + i + 0], z + 0);
                        glVertex3f(x, mHeights[(j + 1) * a + i + 0], z + 1);
                        glEnd();
                    }
                    x++;
                }
                z++;
            }
        } // Draw north-south cliffs

    }
    glPopAttrib();
}

float cPadmap::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
    //return 0;

    float localpos_[3];

    {
        float ori_inv[4];
        quat_cpy(ori_inv, traceable->ori);
        quat_conj(ori_inv);

        vector_cpy(localpos_, worldpos);
        vector_sub(localpos_, localpos_, traceable->pos);
        quat_apply(localpos_, ori_inv, localpos_);
    }

    int a = 16;
    int b = 16;

    float u = (localpos_[0]);
    float v = (localpos_[2]);

    if (u < 0) return 0;
    if (v < 0) return 0;
    if (u >= a / 2) return 0;
    if (v >= b / 2) return 0;

    int u0 = int(u) * 2;
    float du = u - int(u);

    int v0 = int(v) * 2;
    float dv = v - int(v);

    //cout << u0 << " " << v0 << " / " << du << " " << dv << endl;

    float h00 = mHeights[(v0 + 0) * 16 + u0 + 0];
    float h10 = mHeights[(v0 + 1) * 16 + u0 + 0];
    float h11 = mHeights[(v0 + 1) * 16 + u0 + 1];
    float h01 = mHeights[(v0 + 0) * 16 + u0 + 1];

    float h0 = (1 - dv) * h00 + dv * h10;
    float h1 = (1 - dv) * h01 + dv * h11;
    float h = (1 - du) * h0 + du * h1;

    //cout << h00 << " " << h10 << " " << h11 << " " << h01 << endl;
    //cout << h << endl;

    //return 0;

    float localprj[3];
    float depth = 0;
    vector_cpy(localprj, localpos_);
    if (localprj[1] < h + radius) {
        depth = h + radius - localprj[1];
        localprj[1] = h + radius;
    }

    if (depth <= 0) return 0;

    if (localpos != NULL) vector_cpy(localpos, localprj);

    {
        quat_apply(worldpos, traceable->ori, localprj);
        vector_add(worldpos, worldpos, traceable->pos);
    }

    return depth;
}


// ###############################################

cHeightmap::cHeightmap(float x, float z) {
    vector_set(traceable->pos, x, 0, z);

    collideable = new rCollideable;
    addRole(COLLIDEABLE);

    mDims[0] = 20;
    mDims[1] = 20;

    traceable->radius = 24;

    mHeights = new float[mDims[0] * mDims[1]];
    assert(mHeights != NULL);

    loopj(mDims[1]) {

        loopi(mDims[0]) {
            float a = sin(float(i) / float(mDims[0] - 1) * 1 * M_PI);
            float b = sin(float(j) / float(mDims[1] - 1) * 1 * M_PI);
            float h1 = (a * b);

            float c = sin(float(i) / float(mDims[0] - 1) * 1 * M_PI - 0.4 * M_PI);
            float d = sin(float(j) / float(mDims[1] - 1) * 0.7 * M_PI);
            float h2 = 1 - 1.09 * (c * d);
            //mHeights[j * mDims[0] + i] = h1  * 4.2f;
            mHeights[j * mDims[0] + i] = xmin(h1, h2) * 5 - 0.3;
            //mHeights[j * mDims[0] + i] = mDims[0] -1 - i - 0.1;
            //mHeights[j * mDims[0] + i] = i;
            //mHeights[j * mDims[0] + i] = xmin(mDims[0] -1 - i, i);
        }
    }

    mPlanes = new float[mDims[0] * mDims[1] * 8];
    assert(mPlanes != NULL);

    loopj(mDims[1] - 1) {

        loopi(mDims[0] - 1) {
            // The four adjacent vertex heights forming two triangles.
            float h00 = mHeights[(j + 0) * mDims[0] + i + 0];
            float h01 = mHeights[(j + 1) * mDims[0] + i + 0];
            float h11 = mHeights[(j + 1) * mDims[0] + i + 1];
            float h10 = mHeights[(j + 0) * mDims[0] + i + 1];
            float p[3];
            float dx[3];
            float dz[3];
            float* dy;
            // upper triangle
            vector_set(dx, 1, h10 - h00, 0);
            vector_set(dz, 0, h01 - h00, 1);
            dy = &mPlanes[(j * mDims[0] + i) * 8];
            vector_cross(dy, dx, dz);
            vector_norm(dy, dy);
            vector_neg(dy, dy);
            //vector_print(dy);
            vector_set(p, i, h00, j);
            dy[3] = -vector_dot(dy, p);
            //cout << "upper " << dy[3] << endl;
            // lower triangle
            vector_set(dx, -1, h01 - h11, 0);
            vector_set(dz, 0, h10 - h11, -1);
            dy = &dy[4];
            vector_cross(dy, dx, dz);
            vector_norm(dy, dy);
            vector_neg(dy, dy);
            //vector_print(dy);
            vector_set(p, i + 1, h11, j + 1);
            dy[3] = -vector_dot(dy, p);
            //cout << "lower " << dy[3] << endl;
        }
        //cout << endl;
    }
}

float cHeightmap::constrainParticle(float* worldpos, float radius, float* localpos, cObject* enactor) {
    float localpos_[3];

    {
        float ori_inv[4];
        quat_cpy(ori_inv, traceable->ori);
        quat_conj(ori_inv);

        vector_cpy(localpos_, worldpos);
        vector_sub(localpos_, localpos_, traceable->pos);
        quat_apply(localpos_, ori_inv, localpos_);
    }

    float u = (localpos_[0]);
    float v = (localpos_[2]);

    int u0 = floor(u);
    float du = u - u0;

    int v0 = floor(v);
    float dv = v - v0;

    if (u0 < 0 || u0 >= mDims[0] - 1) return 0;
    if (v0 < 0 || v0 >= mDims[1] - 1) return 0;

    //cout << u0 << " " << v0 << " / " << du << " " << dv << endl;
    //cout << u << " " << v <<  endl;

    float* plane = &mPlanes[(v0 * mDims[0] + u0) * 8];
    if (dv > du)
        plane = &plane[4];

    float localprj[3];

    float dist = vector_dot(plane, localpos_) + plane[3];
    float diff = dist - radius;
    //cout << dist << " " << diff << " " << (dv > du) << endl;
    float depth = 0;

    if (true)
        if (diff < 0) {
            depth = -diff;
            vector_scale(localprj, plane, depth);
            vector_add(localprj, localprj, localpos_);
        }

    if (depth <= 0) return 0;

    if (localpos != NULL) vector_cpy(localpos, localprj);

    {
        quat_apply(worldpos, traceable->ori, localprj);
        vector_add(worldpos, worldpos, traceable->pos);
    }

    return depth;
}

/*
float cHeightmap::constrainParticle(float* worldpos, float radius, float* localpos) {
    //return 0;

    float localpos_[3];

    {
        float ori_inv[4];
        quat_cpy(ori_inv, mOri);
        quat_conj(ori_inv);

        vector_cpy(localpos_, worldpos);
        vector_sub(localpos_, localpos_, mPos);
        quat_apply(localpos_, ori_inv, localpos_);
    }

    float u = (localpos_[0]);
    float v = (localpos_[2]);

    if (u < 0) return 0;
    if (v < 0) return 0;
    if (u >= mDims[0]) return 0;
    if (v >= mDims[1]) return 0;

    int u0 = int(u);
    float du = u - int(u);

    int v0 = int(v);
    float dv = v - int(v);

    //cout << u0 << " " << v0 << " / " << du << " " << dv << endl;

    int w = mDims[0];
    float h00 = mHeights[(v0 + 0) * w + u0 + 0];
    float h10 = mHeights[(v0 + 1) * w + u0 + 0];
    float h11 = mHeights[(v0 + 1) * w + u0 + 1];
    float h01 = mHeights[(v0 + 0) * w + u0 + 1];

    float h0 = (1 - dv) * h00 + dv * h10;
    float h1 = (1 - dv) * h01 + dv * h11;
    float h = (1 - du) * h0 + du * h1;

    //cout << h00 << " " << h10 << " " << h11 << " " << h01 << endl;
    //cout << h << endl;

    //return 0;

    float localprj[3];
    float depth = 0;
    vector_cpy(localprj, localpos_);
    if (localprj[1] < h + radius) {
        depth = h + radius - localprj[1];
        localprj[1] = h + radius;
    }

    if (depth <= 0) return 0;

    if (localpos != NULL) vector_cpy(localpos, localprj);

    {
        quat_apply(worldpos, mOri, localprj);
        vector_add(worldpos, worldpos, mPos);
    }

    return depth;
}
 */

void cHeightmap::drawSolid() {
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TEXTURE_BIT);
    {
        glDisable(GL_TEXTURE_2D);

        glPushMatrix();
        {
            glTranslatef(traceable->pos[0], traceable->pos[1], traceable->pos[2]);

            int w = mDims[0];
            glBegin(GL_TRIANGLE_STRIP);
            glNormal3f(0, 1, 0);

            loopj(mDims[1] - 1) {

                loopi(w) {
                    float* n0 = &mPlanes[((j + 0) * w + i) * 8];
                    float* n1 = &n0[4];
                    float h0 = mHeights[(j + 0) * w + i];
                    float h1 = mHeights[(j + 1) * w + i];
                    float in0 = fmax(0.7f, fmin(h0, 0.9f));
                    float in1 = fmax(0.7f, fmin(h1, 0.9f));
                    glColor3f(in0, in0 * 0.9, in0 * 0.8);
                    glNormal3fv(n0);
                    glVertex3f(i, h0, j);
                    glColor3f(in1, in1 * 0.9, in1 * 0.8);
                    glNormal3fv(n1);
                    glVertex3f(i, h1, j + 1);
                }
            }
            glEnd();

        }
        glPopMatrix();
    }
    glPopAttrib();
}
