/* 
 * File:     cComcom.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on August 4, 2009, 8:44 PM
 */

#ifndef _CCOMCOM_H
#define	_CCOMCOM_H

class cComputer;
class cComcom;
class cTarcom;
class cWepcom;
class cSyscom;
class cForcom;
class cNavcom;

#include "cObject.h"
#include "cMech.h"

class cComputer {
public:
    virtual void drawHUD() = 0;
    virtual void process(float spf) = 0;
};

/** Communication Computer
 *  encapsulates sending and receiving of messages
 *  and stores these messages.
 *  Orders to comrades may enhance strategic advance.
 */
class cComcom : public cComputer {
protected:
    cObject* mDevice;
    Console* mConsole;
    OID mLastMessage;
public:
    cComcom(cObject* device);
    void process(float spf);
    void drawHUD();
};

/** Targeting Computer,
 * includes Radar, target-selection and -display.
 */
class cTarcom : public cComputer {
public:
    cObject* mDevice;
public:
    cTarcom(cObject* device);
    void process(float spf);
    void drawHUD();
};

/** System Monitoring Computer
 *  includes damage indicators.
 */
class cSyscom : public cComputer {
public:
    cObject* mDevice;
public:
    cSyscom(cObject* device);
    void process(float spf);
    void drawHUD();
};

/** Weapon Computer
 * currently just draws weapon status.
 */
class cWepcom : public cComputer {
public:
    cObject* mDevice;
public:
    cWepcom(cObject* device);
    void process(float spf);
    void drawHUD();
};

/** Forward Computer
 * Forward-looking information overlay.
 */
class cForcom : public cComputer {
public:
    cMech* mDevice;
public:
    cForcom(cMech* device);
    void process(float spf);
    void drawHUD();
};

// Navigation Computer,
// stores mission critical POints of Interest (POIs),
// routes, map data and encapsulates path-finding.

class cNavcom : public cComputer {
public:
    cObject* mDevice;
    int mWaypoint;
    bool mCyclic;
    std::vector< int > mRoute;
    std::vector< std::vector<float> > mPOIs;
public:
    cNavcom(cObject* device);
    void process(float spf);
    void drawHUD();
};


#endif	/* _CCOMCOM_H */

