/**
 * File:     cController.h
 * Project:  LinWarrior 3D
 * Home:     hackcraft.de
 *
 * Created on March 28, 2008, 21:25 PM (1999)
 *
 * AI and autopilot stacked state machine for controlling entities.
 */

#ifndef _CONTROLLER_H
#define _CONTROLLER_H

class cController;

#include "cObject.h"
#include "cComcom.h"

#include <vector>
#include <string>

/**
 * Autopilot/AI object controller, controlls usualy Mechs.
 * Implements an stack-automaton for object behavior control
 * and plan execution.
 */
class cController {
public:
    /// Available opcodes of stack-machine.
    enum Opcodes {
        WAIT, ATTACK, FOLLOW, GOTO, REPEAT, OPCODE_MAX
    };

    bool mEnabled; // Whether control is enabled (autopilot/ai enabled/disabled).
    cObject* mDevice; // Controlled entity cObject.
    std::vector<OID> mStack;

public:
    /// Initialises a en-/disabled controller for the given entity->
    cController(cObject* entity, bool enabled = true);
    virtual ~cController();

    /// Do a single Instruction step as the top frame on the stack says.
    void process(float spf = 1.0f);

    /// Currently there is no hud.
    void drawHUD() {};

    /// Print out State and Stackinformation.
    void printState();

protected:
    std::string getFrameName();
    unsigned int getFrameSizeOf(int opcode);
    unsigned int getFrameSize();

    // Set Parameter value at top of the stack minus offset.
    void setParameter(int offset, OID value);

    // Get Parameter value from top of the stack minus offset.
    OID getParameter(int offset);

    // Push a single value onto the stack.
    void push(OID value);

    // Remove whole instruction/frame from stack.
    void pop();

public:
    // 1. Transition Function
    // 2. State Function

    // Wait until a Timeout occures or the selected event occures:
    // patrol = true  => Enemy Contact
    void pushWaitEvent(long mseconds = -1, bool patrol = true);
    void waitEvent();

    void pushAttackEnemy(OID entity);
    void attackEnemy();

    void pushFollowLeader(OID entity, bool patrol = true);
    void followLeader();

    void pushGotoDestination(float* v, bool patrol = true);
    void gotoDestination();

    void pushRepeatInstructions(int n);
    void repeatInstructions();

    // pushRandomWalk();

    // pushSendWave(OID entity, float radius, int mseconds, int attributes);

    // pushForValue(int n, int offset, int start, int end, int step);

    // Use these from controlled Object (mDevice):
    // Queries - ask for objects
    // Fuzzy Predicates [0,1] - ask for fulfillment
    // Output Actions - ask for taking action
};


#endif
