/* 
 * File:   snippetsgl.h
 * Author: Benjamin Pickhardt
 * Home:   hackcraft.de
 *
 * Created on September 15, 2008, 12:54 PM
 */

#ifndef _SNIPPETSGL_H
#define	_SNIPPETSGL_H


#define glBindTexture3D(texnamevar, min_linear, mag_linear, repeat_x, repeat_y, repeat_z, WIDTH, HEIGHT, DEPTH, texture_ptr) \
{ \
    glGenTextures(1, &texnamevar); \
    glBindTexture(GL_TEXTURE_3D, texnamevar); \
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, min_linear ? GL_LINEAR : GL_NEAREST); \
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, mag_linear ? GL_LINEAR : GL_NEAREST); \
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, repeat_x ? GL_REPEAT : GL_CLAMP); \
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, repeat_y ? GL_REPEAT : GL_CLAMP); \
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, repeat_z ? GL_REPEAT : GL_CLAMP); \
    glTexImage3D(GL_TEXTURE_3D, 0, GL_RGB8, WIDTH, HEIGHT, DEPTH, 0, GL_RGB, GL_UNSIGNED_BYTE, texture_ptr); \
}


#define glBindTexture2D(texnamevar, min_linear, mag_linear, repeat_x, repeat_y, WIDTH, HEIGHT, DEPTH, texture_ptr) \
{ \
    glGenTextures(1, &texnamevar); \
    glBindTexture(GL_TEXTURE_2D, texnamevar); \
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_linear ? GL_LINEAR : GL_NEAREST); \
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_linear ? GL_LINEAR : GL_NEAREST); \
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, repeat_x ? GL_REPEAT : GL_CLAMP); \
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, repeat_y ? GL_REPEAT : GL_CLAMP); \
    if (DEPTH == 3) { \
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, WIDTH, HEIGHT, 0, GL_BGR, GL_UNSIGNED_BYTE, texture_ptr); \
    } else { \
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, WIDTH, HEIGHT, 0, GL_BGRA, GL_UNSIGNED_BYTE, texture_ptr); \
    } \
}


#define glPrint(name, getenum) \
{ \
    int i; \
    glGetIntegerv(getenum, &i); \
    printf("%s = %i\n", name, i); \
}


// Primitives


#define glVertex3fv2f(pnt, u, v) \
{ \
    glTexCoord2f(u,v); \
    glVertex3fv(pnt); \
}

#define glVertex3fv2fv(pnt, tex) \
{ \
    glTexCoord2f(tex[0], tex[1]); \
    glVertex3fv(pnt); \
}

#define glVertex3fv3fv(pnt, tex) \
{ \
    glTexCoord3fv(tex); \
    glVertex3fv(pnt); \
}


#define glNormAxis() \
{ \
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT); \
    glDisable(GL_TEXTURE_2D); \
    glDisable(GL_LIGHTING); \
    glBegin(GL_LINES); \
        glColor4f(1,0,0,1); \
        glVertex3f(0,0,0); \
        glVertex3f(1,0,0); \
        glColor4f(0,1,0,1); \
        glVertex3f(0,0,0); \
        glVertex3f(0,1,0); \
        glColor4f(0,0,1,1); \
        glVertex3f(0,0,0); \
        glVertex3f(0,0,1); \
    glEnd(); \
    glPopAttrib(); \
}


#define glAxis(l) \
{ \
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT); \
    glDisable(GL_TEXTURE_2D); \
    glDisable(GL_LIGHTING); \
    glBegin(GL_LINES); \
        glColor4f(1,0,0,1); \
        glVertex3f(0,0,0); \
        glVertex3f(l,0,0); \
        glColor4f(0,1,0,1); \
        glVertex3f(0,0,0); \
        glVertex3f(0,l,0); \
        glColor4f(0,0,1,1); \
        glVertex3f(0,0,0); \
        glVertex3f(0,0,l); \
    glEnd(); \
    glPopAttrib(); \
}


#define glLineSquare(border) \
{ \
    float lo = border; \
    float hi = 1 - border; \
    glBegin(GL_LINE_STRIP); { \
        glVertex3f(lo, lo, 0); \
        glVertex3f(hi, lo, 0); \
        glVertex3f(hi, hi, 0); \
        glVertex3f(lo, hi, 0); \
        glVertex3f(lo, lo, 0); \
    } glEnd(); \
}

#define glDisk(segments, radius) \
{ \
    int n = 2 + (segments); \
    float step = 2*M_PI / (float) n; \
    float a = 0; \
    glBegin(GL_TRIANGLE_FAN); \
    glNormal3f(0,0,1); \
    loopi(n) { \
        glTexCoord3f(0.45+0.499*sin(a), 0.5+0.499*cos(a), 0); \
        glVertex3f(sin(a)*(radius), cos(a)*(radius), 0); \
        a += step; \
    } \
    glEnd(); \
}


#define glUnitCylinder(segments) \
{ \
    int n = 2 + (segments); \
    float step = 1.0f / (float) n; \
    float anglestep = 2.0f * M_PI *step; \
    float angle = 0.0f; \
    glBegin(GL_TRIANGLE_STRIP); \
    for (int i=0; i<=n; i++) { \
        glTexCoord3f(0.45+0.499*sin(angle), 0.999f, 0.5+0.499*cos(angle)); \
        glVertex3f(sin(angle), +1.0f, cos(angle)); \
        glTexCoord3f(0.5+0.0.499*sin(angle), 0.001f, 0.5+0.499*cos(angle)); \
        glVertex3f(sin(angle), +0.0f, cos(angle)); \
        angle += anglestep; \
    }; \
    glEnd(); \
}


#define glCenterUnitCylinder(segments) \
{ \
    float step = 1.0f / (float) segments; \
    float anglestep = 2.0f * M_PI *step; \
    float angle = 0.0f; \
    glBegin(GL_TRIANGLE_STRIP); \
    for (int i=0; i<=segments; i++) { \
        glTexCoord3f(0.5+0.499*sin(angle), 0.999f, 0.5+0.499*cos(angle)); \
        glVertex3f(sin(angle), +1.0f, cos(angle)); \
        glTexCoord3f(0.5+0.499*sin(angle), 0.001f, 0.5+0.499*cos(angle)); \
        glVertex3f(sin(angle), -1.0f, cos(angle)); \
        angle += anglestep; \
    }; \
    glEnd(); \
}


#define glUnitBlock() \
{ \
    GLfloat p[][3] = { \
        {0,0,0}, {1,0,0}, {1,1,0}, {0,1,0}, \
        {0,0,1}, {1,0,1}, {1,1,1}, {0,1,1} \
    }; \
    const float a = 0.01f; \
    const float b = 0.99f; \
    GLfloat t[][3] = { \
        {a,a,a}, {b,a,a}, {b,b,a}, {a,b,a}, \
        {a,a,b}, {b,a,b}, {b,b,b}, {a,b,b} \
    }; \
     \
    glBegin(GL_QUADS); \
        /* front */ \
        glNormal3f(0, 0, -1); \
        glVertex3fv3fv(p[1], t[1]); glVertex3fv3fv(p[0], t[0]); \
        glVertex3fv3fv(p[3], t[3]); glVertex3fv3fv(p[2], t[2]); \
        /* back */ \
        glNormal3f(0, 0, +1); \
        glVertex3fv3fv(p[4], t[4]); glVertex3fv3fv(p[5], t[5]); \
        glVertex3fv3fv(p[6], t[6]); glVertex3fv3fv(p[7], t[7]); \
        /* top */ \
        glNormal3f(0, -1, 0); \
        glVertex3fv3fv(p[0], t[0]); glVertex3fv3fv(p[1], t[1]); \
        glVertex3fv3fv(p[5], t[5]); glVertex3fv3fv(p[4], t[4]); \
        /* bottom */ \
        glNormal3f(0, +1, 0); \
        glVertex3fv3fv(p[2], t[2]); glVertex3fv3fv(p[3], t[3]); \
        glVertex3fv3fv(p[7], t[7]); glVertex3fv3fv(p[6], t[6]); \
        /* left */ \
        glNormal3f(+1, 0, 0); \
        glVertex3fv3fv(p[0], t[0]); glVertex3fv3fv(p[4], t[4]); \
        glVertex3fv3fv(p[7], t[7]); glVertex3fv3fv(p[3], t[3]); \
        /* right */ \
        glNormal3f(-1, 0, 0); \
        glVertex3fv3fv(p[5], t[5]); glVertex3fv3fv(p[1], t[1]); \
        glVertex3fv3fv(p[2], t[2]); glVertex3fv3fv(p[6], t[6]); \
    glEnd(); \
}


#define glBlockFlat(width, height, depth, left,right, front,back, top,bottom) \
{ \
    float a = 0, b = 0, c = 0; \
    float x = width; \
    float y = height; \
    float z = depth; \
    GLfloat p[][3] = { \
        {a,b,c}, {x,b,c}, {x,y,c}, {a,y,c}, \
        {a,b,z}, {x,b,z}, {x,y,z}, {a,y,z} \
    }; \
    GLfloat t[][2] = { \
        {a,b}, {x,b}, {x,y}, {a,y}, \
        {c,b}, {z,b}, {z,y}, {c,y}, \
        {a,c}, {x,c}, {x,z}, {a,z}, \
    }; \
    \
    glBindTexture(GL_TEXTURE_2D, front); \
    glBegin(GL_QUADS); \
        glNormal3f(0, 0, -1); \
        glVertex3fv2fv(p[1], t[1]); glVertex3fv2fv(p[0], t[0]); \
        glVertex3fv2fv(p[3], t[3]); glVertex3fv2fv(p[2], t[2]); \
    glEnd(); \
    if (back != front) glBindTexture(GL_TEXTURE_2D, back); \
    glBegin(GL_QUADS); \
        glNormal3f(0, 0, +1); \
        glVertex3fv2fv(p[4], t[1]); glVertex3fv2fv(p[5], t[0]); \
        glVertex3fv2fv(p[6], t[3]); glVertex3fv2fv(p[7], t[2]); \
    glEnd(); \
    if (left != back) glBindTexture(GL_TEXTURE_2D, left); \
    glBegin(GL_QUADS); \
        glNormal3f(+1, 0, 0); \
        glVertex3fv2fv(p[0], t[5]); glVertex3fv2fv(p[4], t[4]); \
        glVertex3fv2fv(p[7], t[7]); glVertex3fv2fv(p[3], t[6]); \
    glEnd(); \
    if (right != left) glBindTexture(GL_TEXTURE_2D, right); \
    glBegin(GL_QUADS); \
        glNormal3f(-1, 0, 0); \
        glVertex3fv2fv(p[5], t[5]); glVertex3fv2fv(p[1], t[4]); \
        glVertex3fv2fv(p[2], t[7]); glVertex3fv2fv(p[6], t[6]); \
    glEnd(); \
    if (top != right) glBindTexture(GL_TEXTURE_2D, top); \
    glBegin(GL_QUADS); \
        glNormal3f(0, -1, 0); \
        glVertex3fv2fv(p[0], t[9]); glVertex3fv2fv(p[1], t[8]); \
        glVertex3fv2fv(p[5], t[11]); glVertex3fv2fv(p[4], t[10]); \
    glEnd(); \
    if (bottom != top) glBindTexture(GL_TEXTURE_2D, bottom); \
    glBegin(GL_QUADS); \
        glNormal3f(0, +1, 0); \
        glVertex3fv2fv(p[2], t[9]); glVertex3fv2fv(p[3], t[8]); \
        glVertex3fv2fv(p[7], t[11]); glVertex3fv2fv(p[6], t[10]); \
    glEnd(); \
}


#define glCenterUnitBlock() \
{ \
    GLfloat p[][3] = { \
        {-1,-1,-1}, {+1,-1,-1}, {+1,+1,-1}, {-1,+1,-1}, \
        {-1,-1,+1}, {+1,-1,+1}, {+1,+1,+1}, {-1,+1,+1} \
    }; \
    const float a = 0.01f; \
    const float b = 0.99f; \
    GLfloat t[][3] = { \
        {a,a,a}, {b,a,a}, {b,b,a}, {a,b,a}, \
        {a,a,b}, {b,a,b}, {b,b,b}, {a,b,b} \
    }; \
     \
    glBegin(GL_QUADS); \
        /* front */ \
        glNormal3f(0, 0, -1); \
        glVertex3fv3fv(p[1], t[1]); glVertex3fv3fv(p[0], t[0]); \
        glVertex3fv3fv(p[3], t[3]); glVertex3fv3fv(p[2], t[2]); \
        /* back */ \
        glNormal3f(0, 0, +1); \
        glVertex3fv3fv(p[4], t[4]); glVertex3fv3fv(p[5], t[5]); \
        glVertex3fv3fv(p[6], t[6]); glVertex3fv3fv(p[7], t[7]); \
        /* top */ \
        glNormal3f(0, -1, 0); \
        glVertex3fv3fv(p[0], t[0]); glVertex3fv3fv(p[1], t[1]); \
        glVertex3fv3fv(p[5], t[5]); glVertex3fv3fv(p[4], t[4]); \
        /* bottom */ \
        glNormal3f(0, +1, 0); \
        glVertex3fv3fv(p[2], t[2]); glVertex3fv3fv(p[3], t[3]); \
        glVertex3fv3fv(p[7], t[7]); glVertex3fv3fv(p[6], t[6]); \
        /* left */ \
        glNormal3f(+1, 0, 0); \
        glVertex3fv3fv(p[0], t[0]); glVertex3fv3fv(p[4], t[4]); \
        glVertex3fv3fv(p[7], t[7]); glVertex3fv3fv(p[3], t[3]); \
        /* right */ \
        glNormal3f(-1, 0, 0); \
        glVertex3fv3fv(p[5], t[5]); glVertex3fv3fv(p[1], t[1]); \
        glVertex3fv3fv(p[2], t[2]); glVertex3fv3fv(p[6], t[6]); \
    glEnd(); \
}


// Pushing and Poping different Projection-Matrices:
// Note that the current Matrix-Mode is saved and
// stays the same after calling the projection-macros (most often MODELVIEW).

#define glPushPerspectiveProjection(fovdegrees) \
{ \
    glPushAttrib(GL_DEPTH_BUFFER_BIT); /* push glDepthFunc */ \
    glPushAttrib(GL_TRANSFORM_BIT); \
        glMatrixMode(GL_PROJECTION); \
        glPushMatrix(); \
        glLoadIdentity(); \
        GLint view[4]; \
        glGetIntegerv(GL_VIEWPORT, view); \
        gluPerspective(fovdegrees, GLdouble(view[2]) / GLdouble(view[3]), 0.07, 1500.0); \
        glDepthFunc(GL_LEQUAL); \
    glPopAttrib(); \
}


#define glPushUnitOrthoProjection() \
{ \
    glPushAttrib(GL_DEPTH_BUFFER_BIT); /* push glDepthFunc */ \
    glPushAttrib(GL_TRANSFORM_BIT); \
        glMatrixMode(GL_PROJECTION); \
        glPushMatrix(); \
        glLoadIdentity(); \
        glOrtho(0, 1, 0, 1, -500.0, 500.0); \
        glDepthFunc(GL_GEQUAL); \
    glPopAttrib(); \
}


#define glPushCenterUnitOrthoProjection() \
{ \
    glPushAttrib(GL_DEPTH_BUFFER_BIT); /* push glDepthFunc */ \
    glPushAttrib(GL_TRANSFORM_BIT); \
        glMatrixMode(GL_PROJECTION); \
        glPushMatrix(); \
        glLoadIdentity(); \
        glOrtho(-1, 1, -1, 1, -500.0, 500.0); \
        glDepthFunc(GL_GEQUAL); \
    glPopAttrib(); \
}


#define glPushViewportOrthoProjection() \
{ \
    glPushAttrib(GL_DEPTH_BUFFER_BIT); /* push glDepthFunc */ \
    glPushAttrib(GL_TRANSFORM_BIT); \
        glMatrixMode(GL_PROJECTION); \
        glPushMatrix(); \
        glLoadIdentity(); \
        GLint view[4]; \
        glGetIntegerv(GL_VIEWPORT, view); \
        glOrtho(view[0], view[2], view[1], view[3], -500.0, 500.0); \
        glDepthFunc(GL_GEQUAL); \
    glPopAttrib(); \
}


#define glPopProjection() \
{ \
    glPushAttrib(GL_TRANSFORM_BIT); \
        glMatrixMode(GL_PROJECTION); \
        glPopMatrix(); \
    glPopAttrib(); \
    glPopAttrib(); /* pop glDepthFunc */ \
}


// Special Screen Effects

#define glAccumBlur(alpha) \
{ \
    glAccum(GL_MULT, alpha); /* accum = accum * factor */ \
    glAccum(GL_ACCUM, 1.0f-alpha); /* accum = accum + color * factor */ \
    glAccum(GL_RETURN, 1.0f); /* color = accum * factor */ \
}


#define glAccumBlurInverse(alpha) \
{ \
    glAccum(GL_MULT, alpha); /* accum = accum * factor */ \
    glAccum(GL_ACCUM, 1.0f-alpha); /* accum = accum + color * factor */ \
    glAccum(GL_MULT, -1.0f); glAccum(GL_ADD, 1.0f); /* inversion */ \
    glAccum(GL_RETURN, 1.0f); /* color = accum * factor */ \
    glAccum(GL_ADD, -1.0f); glAccum(GL_MULT, -1.0f); /* inversion */ \
}


// Billboarding
// Extracts and inverts the rotational part of the MODELVIEW_MATRIX
// by embedding the transposed upper 3x3 into an otherwise
// identity 4x4 matrix - assumes that no scaling was applied.
// The result can be glMultMatrix'ed in to eliminate rotations
// for billboarding.

#define glGetTransposeInverseRotationMatrix(invm) \
{ \
    float _m_[16]; \
    glGetFloatv(GL_MODELVIEW_MATRIX, _m_); \
    float _n_[] = { /* = m Transposed = m Inverted */  \
        _m_[0], _m_[4], _m_[ 8],  0, \
        _m_[1], _m_[5], _m_[ 9],  0, \
        _m_[2], _m_[6], _m_[10],  0, \
             0,      0,       0,  1 \
    }; \
    memcpy(invm, _n_, sizeof(float)*16); \
}


// Extracts and inverts the translational part of the MODELVIEW_MATRIX
// by embedding the inverted lower 1x3 translational vector
// into an otherwise identity 4x4 matrix.

#define glGetInverseTranslationMatrix(invm) \
{ \
    float _m_[16]; \
    glGetFloatv(GL_MODELVIEW_MATRIX, _m_); \
    float _n_[] = { /* = m Transposed = m Inverted */  \
        1, 0, 0,  0, \
        0, 1, 0,  0, \
        0, 0, 1,  0, \
       -_m_[12],-_m_[13],-_m_[14],  1 \
    }; \
    memcpy(invm, _n_, sizeof(float)*16); \
}


// Multiplies a quaternion rotation into the current matrix.

#define glRotateq(quat4fv) \
{ \
    float x = -(quat4fv)[X]; \
    float y = -(quat4fv)[Y]; \
    float z = -(quat4fv)[Z]; \
    float w =  (quat4fv)[W]; \
    float x2 = x*x; \
    float y2 = y*y; \
    float z2 = z*z; \
    float m[] = { \
        1 - 2 * y2 - 2 * z2, 2 * x * y - 2 * w*z, 2 * x * z + 2 * w*y, 0, \
        2 * x * y + 2 * w*z, 1 - 2 * x2 - 2 * z2, 2 * y * z - 2 * w*x, 0, \
        2 * x * z - 2 * w*y, 2 * y * z + 2 * w*x, 1 - 2 * x2 - 2 * y2, 0, \
        0, 0, 0, 1 \
    }; \
    glMultMatrixf(m); \
}


#endif	/* _SNIPPETSGL_H */

